/* bpfilt.c

 * Copyright (C) 2000 Claudio Girardi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * References:
 *
 * G. Var Rosenbaum, Spectrum Astro,
 *   "RF Capacitive-coupled filters",
 *   Applied Microwave and Wireless, October 1998
 *
 * G. Matthaei , L. Young and E. M. T. Jones,
 *   "Microwave filters, impedance-matching networks, and coupling structures",
 *   McGraw Hill, 1964
 *
 * S. B. Cohn,
 *   "Direct-coupled-resonator filters",
 *   Proceedings of the IRE, February 1957
 *
 * S. B. Cohn,
 *   "Dissipation loss in multiple-coupled resonator filters",
 *   Proceedings of the IRE, August 1959
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <math.h>

#ifndef PI
#ifndef M_PI
#define M_PI 3.141592654
#endif
#define PI M_PI
#endif

#define ALOG(x) (pow(10.0, (x)))

#define NMAX 40
#define VERSION "0.1alpha"

static float Am, r, f0, f1, f2, w0, bw, Cn, Ln, Qbp, Q, Qu, R, L0;
static float Rs, Rl;
static float Cs[NMAX], Cp[NMAX], Ls[NMAX], Lp[NMAX], g[NMAX],
 q[NMAX], k[NMAX], K[NMAX];
static int i, n;
static int filt_type;
static char coupling_type, out_format, matching_type;

static char *filt_name[] = {
  "Equiripple [Tchebyscheff]",
  "Maximally flat [Butterworth]",
  "Minimum loss [Cohn]"
};

static char *program_name;
static FILE *fout;
static char fout_name[20];


static void usage(void)
{
  int i = 0;

  fprintf(stderr,
	  "%s, by IN3OTD, %s, Version "VERSION"\n"	    
	  "Usage: %s [OPTIONS]\n",
	  program_name, __DATE__, program_name);
  fprintf(stderr,
	  "options: -c coupling_type : c = capacitive\n"
	  "                            l = inductive\n"
	  "         -f output_format : s = Spice\n"
	  "                            a = Aplac\n"
	  "         -h               : print this help\n"
	  "         -m matching_type : s = series cap\n"
	  "                            p = parallel cap\n"
	  "         -o filename      : output file name\n"
	  "         -t filter_type   : %i = %s\n"
	  "         -v               : print version informations\n"
	  , i, filt_name[i]);

  for(i = 1; i < 3; i++)
    fprintf(stderr,
	    "                            %i = %s\n",
	    i, filt_name[i]);
	  fprintf(stderr,"\n");
	  fprintf(stderr,"Report bugs to <in3otd@qsl.net>\n");
	  fprintf(stderr,"\n");
}


static void version(void)
{
  fprintf(stderr,
	  "bpfilt "VERSION"\n"
	  "Copyright (C) 2000 Claudio Girardi\n"
	  "This program is distributed in the hope that it will be useful,\n"
	  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	  "GNU General Public License for more details.\n"
	  "\n");
}


static void initialize(void)
{
  int i;

  /* open output file */
  if (!(fout = fopen(fout_name, "w"))) {
    fprintf(stderr, "Error opening output file!\n");
    exit(-1);
  }

  /* clear component values */
  for (i = 0; i < NMAX; i++) {
    Cp[i] = 0.0;
    Cs[i] = 0.0;
    Lp[i] = 0.0;
    Ls[i] = 0.0;
  }
}


static void write_aplac_file(void)
{
  time_t date;

  time(&date);

  fprintf(fout, "$  ---------------------------\n");
  fprintf(fout, "$  File     : %s\n", fout_name);
  fprintf(fout, "$  LastEdit : %.24s\n", ctime(&date));
  fprintf(fout, "$  ---------------------------\n");
  fprintf(fout, "\n");
  fprintf(fout, "$  Bandpass filter (%s)\n", filt_name[filt_type]);
  fprintf(fout, "$    Center Freq. : %e Hz\n", f0);
  fprintf(fout, "$    Bandwidth    : %e Hz\n", bw);
  fprintf(fout, "$    Passband Ripple : %e dB\n", Am);
  fprintf(fout, "$    Source Impedance : %e ohm\n", Rs);
  fprintf(fout, "$    Load Impedance   : %e ohm\n", Rl);
  fprintf(fout, "$    %i Resonator(s)\n", n);
  fprintf(fout, "$      Node Inductance  : %e H\n", Ln);
  fprintf(fout, "$      Node Capacitance : %e F\n", Cn);
  fprintf(fout, "\n");
  fprintf(fout, "\n");

  if (Cp[0] != 0.0)
    fprintf(fout, "Cap Cp0 in 0  %e\n", Cp[0]);
  if (Cs[0] != 0.0)
    fprintf(fout, "Cap Cs0 in 1  %e\n", Cs[0]);
  for (i = 1; i <= n; i++) {
    if (Cp[i] != 0.0)
      fprintf(fout, "Cap Cp%-3i %i 0  %e\n", i, i, Cp[i]);
    if (Lp[i] != 0.0)
      fprintf(fout, "Ind Lp%-3i %i 0  %e G=%e\n", i, i, Lp[i], 1 / (Qu * w0 * Lp[i]));
    if (Ls[i] != 0.0)
      fprintf(fout, "Ind Ls%-3i %i %i  %e\n", i, i, i + 1, Ls[i]);
    if (Cs[i] != 0.0)
      if (i != n) {
	fprintf(fout, "Cap Cs%-3i %i %i  %e\n", i, i, i + 1, Cs[i]);
      } else {
	fprintf(fout, "Cap Cs%-3i %i out  %e\n", i, i, Cs[i]);
      }
  }
  if (Cp[n + 1] != 0.0)
    fprintf(fout, "Cap Cp%-3i out 0  %e\n", n + 1, Cp[n + 1]);

  fprintf(fout, "\n");
  fprintf(fout, "\n");
  fprintf(fout, "DefNPort Filt 2 in 0 %f out 0 %f\n", Rs, Rl);
  fprintf(fout, "\n");
  fprintf(fout, "Sweep Filt\n");
  fprintf(fout, "+ AC\n");
  fprintf(fout, "+ LOOP 151 FREQ LIN %e %e\n", f0 / 1.2, f0 * 1.2);
  //fprintf(fout, "+ LOOP 151 FREQ LIN %e %e\n", 120e3, 160e3);
  fprintf(fout, "+ NWINDOW 0\n");
  fprintf(fout, "+ BIG_SCREEN\n");
  fprintf(fout, "+ SMITH\n");
  fprintf(fout, "+ NWINDOW 1\n");
  fprintf(fout, "+ GRID\n");
  fprintf(fout, "  Show WINDOW 0 RI S(1,1)\n");
  //fprintf(fout, "+ CSDF topo%c%c\n", matching_type, coupling_type);
  //fprintf(fout, "  Show WINDOW 1 DB S(2,1)\n");
  fprintf(fout, "  +    WINDOW 1 DB S(2,1)\n");
  fprintf(fout, "  +    WINDOW 1 DB Mag(S(1,1))\n");
  fprintf(fout, "EndSweep\n");

}


static void write_spice_file(void)
{
  time_t date;

  time(&date);
  fprintf(fout, "*  Bandpass filter (%s)\n", filt_name[filt_type]);
  fprintf(fout, "*  File     : %s\n", fout_name);
  fprintf(fout, "*  LastEdit : %.24s\n", ctime(&date));
  fprintf(fout, "*\n");
  fprintf(fout, "*\n");
  fprintf(fout, "*    Center Freq. : %e Hz\n", f0);
  fprintf(fout, "*    Bandwidth    : %e Hz\n", bw);
  fprintf(fout, "*    Passband Ripple : %e dB\n", Am);
  fprintf(fout, "*    Source Impedance : %e ohm\n", Rs);
  fprintf(fout, "*    Load Impedance   : %e ohm\n", Rl);
  fprintf(fout, "*    %i Resonator(s)\n", n);
  fprintf(fout, "*      Node Inductance  : %e H\n", Ln);
  fprintf(fout, "*      Node Capacitance : %e F\n", Cn);
  fprintf(fout, "\n");
  fprintf(fout, "\n");

  fprintf(fout, "Vin inp 0  0 AC 1\n");
  fprintf(fout, "Ri  inp in %f\n", Rs);
  if (Cp[0] != 0.0)
    fprintf(fout, "Cp0 in 0  %e\n", Cp[0]);
  if (Cs[0] != 0.0)
    fprintf(fout, "Cs0 in 1  %e\n", Cs[0]);
  for (i = 1; i <= n; i++) {
    if (Cp[i] != 0.0)
      fprintf(fout, "Cp%-3i %i 0  %e\n", i, i, Cp[i]);
    if (Lp[i] != 0.0) {
      fprintf(fout, "Lp%-3i %i %ia  %e\n", i, i, i, Lp[i]);
      fprintf(fout, "Rp%-3i %ia 0  %e\n", i, i, (w0 * Lp[i] / Qu));
    }
    if (Ls[i] != 0.0) {
      fprintf(fout, "Ls%-3i %i %i  %e\n", i, i, i + 1, Ls[i]);
    }
    if (Cs[i] != 0.0)
      if (i != n) {
	fprintf(fout, "Cs%-3i %i %i  %e\n", i, i, i + 1, Cs[i]);
      } else {
	fprintf(fout, "Cs%-3i %i out  %e\n", i, i, Cs[i]);
      }
  }
  if (Cp[n + 1] != 0.0)
    fprintf(fout, "Cp%-3i out 0  %e\n", n + 1, Cp[n + 1]);
  fprintf(fout, "Ro  out 0 %f\n", Rl);

  fprintf(fout, "\n");
  fprintf(fout, "\n");
  fprintf(fout, ".CONTROL\n");
  fprintf(fout, "  ac lin 151 %e %e\n", f0 / 1.2, f0 * 1.2);
  fprintf(fout, "  plot db(out)%+f\n", 10.0 * log10(4.0 * Rs / Rl));
  fprintf(fout, ".ENDC\n");

}


static void compute_tchebyscheff(void)
{
  static float wp;
  static float beta, nu, a[NMAX], b[NMAX];
  static float A=1.0;

  wp = cosh(acosh(sqrt((ALOG(Am / 10.0) - 1.0) / (ALOG(A / 10.0) - 1.0))) / n);
  bw = bw / wp;
#ifdef DEBUG
  printf("3dB bw = %f\n", bw);
#endif
  Am = 1.0;

  beta = log(1 / (tanh(Am / 17.37)));
  nu = sinh(beta / (2 * n));

  for (i = 1; i < n + 1; i++) {
    a[i] = sin((2 * i - 1) * PI / (2 * n));	/* a[1] ... a[n] */
    b[i] = nu * nu + sin(i * PI / n) * sin(i * PI / n);	/* b[1] ... b[n] */
  }

  if ((n & 1) == 0) {
    r = tanh(beta / 4) * tanh(beta / 4);	/* n even */
  } else {
    r = 1;			/* n odd */
  }

  g[0] = 1;			/* g[0] */
  g[1] = 2 * a[1] / nu;		/* g[1] */

  for (i = 2; i <= n; i++) {	/* g[2] ... g[n] */
    g[i] = 4 * a[i - 1] * a[i] / (b[i - 1] * g[i - 1]);
  }
}


static void compute_cohn(void)
{

  float g_cohn;

  r = 1.0;
  /* compute the -3dB bandwidth using the stopband approximation */
  bw = bw / pow(10, (Am + 10 * log10(4.0 * r)) / (20.0 * n));
  g[0] = 1.0;
  g_cohn = 1.0;
  for (i = 1; i <= n; i++) {	/* g[1] ... g[n] */
    g[i] = g_cohn;
  }
}


static void compute_max_flat(void)
{

  float wp;

  r = 1.0;
  wp = pow((pow(10, Am / 10.0) - 1.0), (1.0 / (2.0 * n)));
  bw = bw / wp;
#ifdef DEBUG
  printf("3dB bw = %f\n", bw);
#endif
  g[0] = 1.0;
  for (i = 1; i <= n; i++) {	/* g[1] ... g[n] */
    g[i] = 2.0 * sin((2 * i - 1) * PI / (2 * n));
  }
}


static void transform_frequency(void)
{
  //f1 = f0 - bw / 2.0;
  //f2 = f0 + bw / 2.0;

  //f0 = f1 + f2 - sqrt((f2 - f1) * (f2 - f1) + f1 * f2);
  //bw = (f0 / f1 - f0 / f2) * f0;

  float bw2 = bw / 2.0;
  float f03 = f0 / 3.0;

  f1 = -bw2 + 2.0 * f03 + sqrt(bw2 * bw2 + f03 * f03);
  f2 = f1 + bw;
  
  printf("f0 = %f\n", f0);
  printf("f1 = %f\n", f1);
  printf("f2 = %f\n", f2);
  printf("f2-f1 = %f\n", f2-f1);
  printf("bw = %f\n", bw);

}


static void matching_with_series_cap(void)
{
  float Cpi, Cpo, Csi, Cso;

  Csi = sqrt(1 / (w0 * w0 * Rs * (R - Rs)));
  //Csi = 1/(w0*sqrt(q[0]*Rs*R-Rs*Rs));
  Cpi = Csi * (Rs - R) / R;
  //Cpi = Csi/(1+(Rs*w0*Csi)*(Rs*w0*Csi));
  Cso = sqrt(1 / (w0 * w0 * Rl * (R - Rl)));
  //Cso = 1/(w0*sqrt(q[n]*Rl*R-Rl*Rl));
  Cpo = Cso * (Rl - R) / R;
  //Cpo = Cso/(1+(Rl*w0*Cso)*(Rl*w0*Cso));

  Cs[0] = Csi;
  Cp[1] = Cp[1] + Cpi;
  Cs[n] = Cso;
  Cp[n] = Cp[n] + Cpo;
#ifdef DEBUG
  printf("Cs[0] = %e\n", Cs[0]);
  printf("Cp[1] = %e\n", Cp[1]);
  printf("Cs[n] = %e\n", Cs[n]);
  printf("Cp[n] = %e\n", Cp[n]);
#endif
}


static void matching_with_parallel_cap(void)
{
  Cp[0] = sqrt((Rs / R) * (1 + (w0 * Cp[1] * R) * (w0 * Cp[1] * R)) - 1) / (w0 * Rs);
  Cs[0] = (1 + (w0 * Cp[0] * Rs) * (w0 * Cp[0] * Rs)) /
    (w0 * w0 * Rs * (Cp[1] * R - Cp[0] * Rs));
#ifdef DEBUG
  printf("Cp[0] = %e\n", Cp[0]);
  printf("Cs[0] = %e\n", Cs[0]);
#endif

  Cp[n + 1] = (sqrt((Rl / R) * (1 + (w0 * Cp[n] * R) * (w0 * Cp[n] * R)) - 1) /
	      (w0 * Rl));
  Cs[n] = ((1 + (w0 * Cp[n + 1] * Rl) * (w0 * Cp[n + 1] * Rl)) /
	  (w0 * w0 * Rl * (Cp[n] * R - Cp[n + 1] * Rl)));
#ifdef DEBUG
  printf("Cp[n+1] = %e\n", Cp[n + 1]);
  printf("Cs[n] = %e\n", Cs[n]);
#endif
  
  Cp[1] = 0.0;
  Cp[n] = 0.0;
}


static void cap_coupled(void)
{

  for (i = 1; i < n; i++) {
    K[i] = k[i] / Qbp;	/* Coupling Coefficients */
    Cs[i] = K[i] * Cn;		/* Coupling Capacitances */
#ifdef DEBUG
    printf("K[%d,%d]=%f\n", i, i + 1, K[i]);
#endif
  }

  Cp[1] = Cn - Cs[1];
  Lp[1] = Ln;
  for (i = 2; i < n; i++) {	
    Cp[i] = Cn - Cs[i - 1] - Cs[i];  /* Resonator Capacitances */
    Lp[i] = Ln;                 /* Resonator Inductances */
  }
  Cp[n] = Cn - Cs[n - 1];
  Lp[n] = Ln;

#ifdef DEBUG
  for (i = 1; i < n + 1; i++) {
    printf("Cp[%d]=%e\n", i, Cp[i]);
  }

  for (i = 0; i < n + 1; i++) {
    printf("Cs[%d,%d]=%e\n", i, i + 1, Cs[i]);
  }
#endif
}


static void ind_coupled(void)
{
  //Cs1 = sqrt(1 / (w0 * w0 * Rs * (R - Rs)));
  //Cs1 = 1/(w0*sqrt(q[0]*Rs*R-Rs*Rs));
  //Cp1 = Cs1 * (Rs - R) / R;
  //Cp1 = Cs1/(1+(Rs*w0*Cs1)*(Rs*w0*Cs1));
  //Cs2 = sqrt(1 / (w0 * w0 * Rl * (R - Rl)));
  //Cs2 = 1/(w0*sqrt(q[n]*Rl*R-Rl*Rl));
  //Cp2 = Cs2 * (Rl - R) / R;
  //Cp2 = Cs2/(1+(Rl*w0*Cs2)*(Rl*w0*Cs2));

  for (i = 1; i < n; i++) {
    K[i] = k[i] / Qbp;	/* Coupling Coefficients */
    Ls[i] = Ln / K[i];		/* Coupling Inductances */
#ifdef DEBUG
    printf("K[%d,%d]=%f\n", i, i + 1, K[i]);
    printf("Ls[%d,%d]=%f\n", i, i + 1, Ls[i]);
#endif
  }

  Cp[1] = Cn;
  Lp[1] = 1 / ((1 / Ln) - (1 / Ls[1]));
  for (i = 2; i < n; i++) {
    Cp[i] = Cn;			/* Resonator Capacitances */
    Lp[i] = 1 / ((1 / Ln) - (1 / Ls[i - 1]) - (1 / Ls[i]));	/* Resonator Inductances */
  }
  Cp[n] = Cn;
  Lp[n] = 1 / ((1 / Ln) - (1 / Ls[n - 1]));

#ifdef DEBUG
  for (i = 1; i < n + 1; i++) {
    printf("Cp[%d]=%e\n", i, Cp[i]);
    printf("Lp[%d]=%e\n", i, Lp[i]);
  }

  for (i = 0; i < n + 1; i++) {
    printf("Ls[%d,%d]=%e\n", i, i + 1, Ls[i]);
  }
#endif
}


void main(int argc, char *argv[])
{
  int optc;			/* For getopt */
  extern char *optarg;
  extern int optind, opterr, optopt;
  
  /* Construct the name of the executable, without the directory part */
  program_name = strrchr(argv[0], '/');
  if (!program_name)
    program_name = argv[0];
  else
    ++program_name;

   /* Default values */
  coupling_type = 'c';
  out_format = 'a';
  matching_type = 's';
  filt_type = 2;
  strcpy(fout_name, "bpfcir.txt");
  /* Parse the options */
  while ((optc = getopt(argc, argv, "c:f:hm:o:t:v")) != EOF) {
    switch (optc) {
    case 'c': 
      coupling_type = *optarg;
      break;
    case 'f': 
      out_format = *optarg;
      break;
    case 'h': 
      usage();
      exit(0);
      break;
    case 'm': 
      matching_type = *optarg;
      break;
    case 'o': 
      strcpy(fout_name, optarg);
      break;
    case 't': 
      filt_type = atoi(optarg);
      break;
    case 'v': 
      version();
      exit(0);
    default:
      usage();
      exit(0);
    }
  }

  initialize();

  printf("Enter Center Frequency [Hz]    : ");
  scanf("%f", &f0);
  switch (filt_type) {
  case 0:
  case 1:
    printf("Enter Bandwidth [Hz]           : ");
    break;
  case 2:
    printf("Enter Bandwidth [Hz]           : ");
    break;
  default:
    break;
  }
  scanf("%f", &bw);

  switch (filt_type) {
  case 0:
  case 1:
    printf("Enter Passband Ripple [dB]     : ");
    break;
  case 2:
    printf("Enter Stopband Attenuation [dB]: ");
    break;
  default:
    break;
  }
  scanf("%f", &Am);

  printf("Enter Number of Resonators     : ");
  scanf("%d", &n);
  printf("Enter Resonators Unloaded Q    : ");
  scanf("%f", &Qu);
  printf("Enter Inductance L [H]         : ");
  scanf("%f", &Ln);
  printf("Enter Source Impedance [ohm]   : ");
  scanf("%f", &Rs);
  printf("Enter Load Impedance [ohm]     : ");
  scanf("%f", &Rl);

  //transform_frequency();
#ifdef DEBUG
  printf("f0 = %f\n", f0);
  printf("bw = %f\n", bw);
#endif

  switch (filt_type) {
  case 0:
    compute_tchebyscheff();
    break;
  case 1:
    compute_max_flat();
    break;
  case 2:
    compute_cohn();
    break;
  default:
    filt_type = 2;
    compute_cohn();
    break;
  }

  if ((n & 1) == 0) {		/* even vs odd check for g[n+1] */
    g[n + 1] = 1.0 / r;		/* even check */
  } else {
    g[n + 1] = 1.0;		/* odd check */
  }

  w0 = 2 * PI * f0;
  /* here we assume that bw is the -3dB bandwidth */
  Qbp = f0 / bw;
  
  L0 = 0.0;
  for (i = 1; i <= n; i++) {
    L0 += g[i] / Qu;
  }
  L0 = 4.343 * Qbp * L0;
#ifdef DEBUG
  printf("Minimum theoretical loss : %f dB\n", L0);

  for (i = 0; i <= n + 1; i++)
    printf("%.4f ", g[i]);
  printf("\n");
#endif

  q[1] = g[0] * g[1];		/* q[1] */
  q[n] = g[n] * g[n + 1];	/* q[n] */

  for (i = 1; i < n; i++) {
    k[i] = 1 / sqrt(g[i] * g[i + 1]);	/* k[1,2 ... n-1, n] */
  }

  Q = Qbp * q[1];		/* External Q */
  Cn = 1 / (w0 * w0 * Ln);      /* Node capacitance */
  R = w0 * Ln * Q;
#ifdef DEBUG  
  printf("Ext. Q = %f\n", Q);
  printf("Node C = %e\n", Cn);
  printf("R = %f\n", R);
#endif

  if (coupling_type == 'c') {
    cap_coupled();
  } else {
    ind_coupled();
  }

  if (matching_type == 's') {
    matching_with_series_cap();
  } else {
    matching_with_parallel_cap();
  }
  if (out_format == 's') {
    write_spice_file();
  } else {
    write_aplac_file();
  }

  fclose(fout);

}
